"""yojenkins package version defintion"""

__title__ = 'yojenkins'
__summary__ = "A CLI tool to interface with Jenkins server"
__version__ = "0.0.85"
__author__ = "Ismet Handzic"
__copyright__ = "Copyright 2022 %s" % __author__
